/*
 * sun-rbootd   yamamori@kt.rim.or.jp
 */

#ifdef __svr4__
# define SOLARIS2
#else
# define SUNOS4
#endif

#include <strings.h>

#include <sys/file.h>
#include <sys/stropts.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/ioctl.h>

#include <net/if.h>

#ifdef SUNOS4
#include <net/nit_if.h>
#include <net/nit_pf.h>
#include <net/packetfilt.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>
#endif
#ifdef SOLARIS2
#include <stdio.h>
#include <sys/fcntl.h>
#include <sys/sockio.h>
#include <sys/bufmod.h>
#include <sys/pfmod.h>
#include <sys/stropts.h>
#include <sys/ethernet.h>
#include <sys/dlpi.h>
#include <netinet/in.h>
#endif


typedef unsigned char u_int8_t;
typedef unsigned short u_int16_t;

#include "rmp.h"
#undef RMP_MAX_PACKET
#define RMP_MAX_PACKET 1514
#include "rmp_var.h"

static struct ether_header ether_header[0];
static struct rmp_packet rmp_packet[0];

#define OFFSET(str,mem) ((char*)&(str)[0].mem-(char*)str)
#define NIT_OFFSET(str,mem) (OFFSET(str,mem)/sizeof(short))

static const struct ether_addr rmp_addr = {RMP_ADDR};


#define DEV_NIT "/dev/nit"

#define MAXNUMIF 8


char *
ether_interfaces(void)
{
  char cbuf[MAXNUMIF * sizeof (struct ifreq)];
  struct ifconf ifc;
  int sfd;
  int i, j;
  int numinter;
  static char result[IFNAMSIZ];
  static char *const validname[] = {"le", "ie", "ec", 0};


  if (result[0] != '\0') {
    return result;
  }

  if ((sfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
    return 0;
  }
  ifc.ifc_len = sizeof cbuf;
  ifc.ifc_buf = cbuf;
  if (ioctl(sfd, SIOCGIFCONF, (char *)&ifc) < 0) {
    close(sfd);
    return 0;
  }
  close(sfd);

  numinter = ifc.ifc_len / sizeof (struct ifreq);
  for (i = 0; i < numinter; ++i) {
    for (j = 0; validname[j] != 0; ++j) {
      if (strncmp(ifc.ifc_req[i].ifr_name, validname[j], 2) == 0) {
	strcpy(result, ifc.ifc_req[i].ifr_name);
	return result;
      }
    }
  }
  return 0;
}



static int
i_str(int fd, int cmd, char *ptr, int len)
{
  struct strioctl si;

  si.ic_cmd = cmd;
  si.ic_timout = INFTIM;
  si.ic_len = len;
  si.ic_dp = ptr;
  return ioctl(fd, I_STR, (char *)&si);
}


int
ether_open(char *name)
{
  int fd;
  struct ifreq ifr;
  struct packetfilt filter;
  unsigned short *fptr;


#ifdef SUNOS4
  if ((fd = open(DEV_NIT, O_RDWR)) < 0) {
    return -1;
  }
#endif
#ifdef SOLARIS2
  {
    int num;
    char dev[32], *s;

    sprintf(dev, "/dev/%s", name);
    for (s = dev; *s; s++) {
      if ('0' <= *s && *s <= '9') {
	num = atoi(s);
	*s = '\0';
	break;
      }
    }
    if ((fd = open(dev, O_RDWR)) < 0) {
      return -1;
    }
    if (AttachDevice(fd, num) == 0) {
      return -1;
    }
  }
#endif

  if (ioctl(fd, I_SRDOPT, (char *)RMSGD) < 0) {
    return -1;
  }

#ifdef SUNOS4
  strncpy(ifr.ifr_name, name, sizeof ifr.ifr_name);
  if (i_str(fd, NIOCBIND, (char *)&ifr, sizeof ifr) < 0) {
    return -1;
  }
#endif
#ifdef SOLARIS2
  if (BindProtocol(fd, 0, 0, DL_CLDLS, 0, 0) == 0) {
    return -1;
  }
  if (i_str(fd, DLIOCRAW, NULL, 0) < 0) {
    return -1;
  }
#endif


#ifdef SUNOS4
  if (ioctl(fd, I_PUSH, "pf") < 0) {
    return -1;
  }
#endif
#ifdef SOLARIS2
  if (ioctl(fd, I_PUSH, "pfmod") < 0) {
    return -1;
  }
#endif

  fptr = &filter.Pf_Filter[0];

  *fptr++ = ENF_PUSHWORD + NIT_OFFSET(ether_header, ether_type);
  *fptr++ = ENF_PUSHLIT | ENF_LE;
  *fptr++ = htons(1500);
  *fptr++ = ENF_PUSHZERO | ENF_CNOR;

  *fptr++ = ENF_PUSHWORD + NIT_OFFSET(rmp_packet, hp_llc.dsap);
  *fptr++ = ENF_PUSHLIT | ENF_CAND;
  *fptr++ = htons((IEEE_DSAP_HP<<8)|IEEE_SSAP_HP);

  *fptr++ = ENF_PUSHWORD + NIT_OFFSET(rmp_packet, hp_llc.cntrl);
  *fptr++ = ENF_PUSHLIT | ENF_CAND;
  *fptr++ = htons(IEEE_CNTL_HP);

  *fptr++ = ENF_PUSHWORD + NIT_OFFSET(rmp_packet, hp_llc.filler);
  *fptr++ = ENF_PUSHZERO | ENF_CAND;

  *fptr++ = ENF_PUSHWORD + NIT_OFFSET(rmp_packet, hp_llc.dxsap);
  *fptr++ = ENF_PUSHLIT | ENF_CAND;
  *fptr++ = htons(HPEXT_DXSAP);

  *fptr++ = ENF_PUSHWORD + NIT_OFFSET(rmp_packet, hp_llc.sxsap);
  *fptr++ = ENF_PUSHLIT | ENF_CAND;
  *fptr++ = htons(HPEXT_SXSAP);

  filter.Pf_FilterLen = fptr - &filter.Pf_Filter[0];
#ifdef SUNOS4
  if (i_str(fd, NIOCSETF, (char *)&filter, sizeof filter) < 0) {
    return -1;
  }
#endif
#ifdef SOLARIS2
  if (i_str(fd, PFIOCSETF, (char *)&filter, sizeof filter) < 0) {
    return -1;
  }
#endif

#ifdef SUNOS4
  ifr.ifr_addr.sa_family = AF_NIT;
  *(struct ether_addr *)&ifr.ifr_addr.sa_data = rmp_addr;
  if (ioctl(fd, SIOCADDMULTI, &ifr) < 0) {
    return -1;
  }
#endif
#ifdef SOLARIS2
  if (PromMode(fd, DL_PROMISC_PHYS) < 0) {
    return -1;
  }
  if (PromMode(fd, DL_PROMISC_SAP) < 0) {
    return -1;
  }
#endif

  ioctl(fd, I_FLUSH, (char *)FLUSHR);

  return fd;
}


int
ether_write(int fd, char *pkt, int len)
{
#ifdef SUNOS4
  struct sockaddr sa;
  struct strbuf cbuf;
  struct strbuf dbuf;

  sa.sa_family = AF_UNSPEC;
  *(struct ether_header *)sa.sa_data = *(struct ether_header *)pkt;

  cbuf.maxlen = cbuf.len = sizeof sa;
  cbuf.buf = (char *)&sa;

  dbuf.maxlen = dbuf.len = len;
  dbuf.buf = pkt + sizeof (struct ether_header);

  return putmsg(fd, &cbuf, &dbuf, 0);
#endif
#ifdef SOLARIS2
  struct strbuf dbuf;
  dbuf.len = len + sizeof (struct ether_header);
  dbuf.buf = pkt;
  return putmsg(fd, NULL, &dbuf, 0);
#endif
}


int
ether_read(int fd, char *pkt, int len)
{
  return read(fd, pkt, sizeof (struct ether_header) + len);
}


/* --- Following sources are partially taken from "cap60" ---*/

#ifdef SOLARIS2


/*
 * sdlpi.c - Simple "protocol" level interface to Streams based DLPI
 * (SunOS 5.x) (derived from snitp.c SunOS 4.x module)
 *
 *  Provides ability to read/write packets at ethernet level
 *
 *
 * Copyright (c) 1988 by The Trustees of Columbia University 
 *  in the City of New York.
 *
 * Permission is granted to any individual or institution to use,
 * copy, or redistribute this software so long as it is not sold for
 * profit, provided that this notice and the original copyright
 * notices are retained.  Columbia University nor the author make no
 * representations about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied
 * warranty.
 *
 * Edit History:
 *
 *  July 1988  CCKim Created
 *  April '91  djh   Add Phase 2 support
 *  May-June 93  montjoy@thor.ece.uc.EDU,
 *               appro@fy.chalmers.se        SunOS 5.x support
 *
 */

static char columbia_copyright[] = "Copyright (c) 1988 by The Trustees of \
Columbia University in the City of New York";

/*
 * DLPI Support Routines
 *
 */

#define EHRD (sizeof(struct ether_addr))

Acknoledge (dlp_p,ack,msg)
union DL_primitives *dlp_p;
int ack;
char *msg;
{
        if (dlp_p->dl_primitive != ack) {
		fprintf(stderr,"dlpi: %s is nacked.\n",msg);
		if (dlp_p->dl_primitive == DL_ERROR_ACK)
			fprintf(stderr,	"dlpi: dlpi_errno %d\n"
					"dlpi: unix_errno %d\n",
			dlp_p->error_ack.dl_errno,
			dlp_p->error_ack.dl_unix_errno);
	   	else
			fprintf(stderr,"dlpi: spiritual primitive %d.\n",
			dlp_p->dl_primitive);
		return(0);
	}
	return(1);
}

AttachDevice(fd,devno)
int fd,devno;
{
  	int retval;
  	int flags = RS_HIPRI;
  	struct strbuf ctlbuf;
  	union DL_primitives rcvbuf;
  	dl_attach_req_t Request;


  	/* bind to underlying interface */
  	Request.dl_primitive 	= DL_ATTACH_REQ;
  	Request.dl_ppa 		= devno;
  	ctlbuf.len 		= sizeof(Request);
  	ctlbuf.buf 		= (caddr_t)&Request;

  	if (putmsg(fd, &ctlbuf ,NULL,0)  < 0) {
    		perror("Attach Device:");
    		return(0);
  	}

  	ctlbuf.maxlen = sizeof(union DL_primitives);
  	ctlbuf.len = 0;
  	ctlbuf.buf = (char *)&rcvbuf;
  	if ((retval = getmsg(fd, &ctlbuf ,NULL, &flags))  < 0) {
    		perror("Attach Device ack!");
    		return(0);
  	}

	return (Acknoledge(&rcvbuf,DL_OK_ACK,"DL_ATTACH_REQ"));
}

BindProtocol(fd,sap,max_conind,service_mode, conn_mgmt, xidtest_flg )
int	fd,sap,max_conind,service_mode, conn_mgmt, xidtest_flg ;
{
  	int retval;
  	int flags = RS_HIPRI;
  	struct strbuf ctlbuf;
  	union DL_primitives rcvbuf;
  	dl_bind_req_t   BindRequest;


  	BindRequest.dl_primitive 		= DL_BIND_REQ;
  	BindRequest.dl_sap       		= sap;
  	BindRequest.dl_max_conind       	= max_conind;
  	BindRequest.dl_service_mode       	= service_mode;
  	BindRequest.dl_conn_mgmt       		= conn_mgmt;
  	BindRequest.dl_xidtest_flg       	= xidtest_flg;
	
  	ctlbuf.len = sizeof(BindRequest);
  	ctlbuf.buf = (caddr_t)&BindRequest;
	
  	if (putmsg(fd, &ctlbuf ,NULL,0)  < 0) {
    		perror("Bind Protocol:");
    		return(0);
  	}

  	ctlbuf.maxlen = sizeof(union DL_primitives);
  	ctlbuf.len = 0;
  	ctlbuf.buf = (char *)&rcvbuf;
  	if ((retval = getmsg(fd, &ctlbuf ,NULL, &flags))  < 0) {
    		perror("Bind Protocol ACK!");
    		return(0);
  	}

	return (Acknoledge(&rcvbuf,DL_BIND_ACK,"DL_BIND_REQ"));
}

PromMode(fd,level)
int	fd,level;
{
  	int retval;
  	int flags = RS_HIPRI;
  	struct strbuf ctlbuf;
  	union DL_primitives rcvbuf;
	dl_promiscon_req_t PromRequest;


  	PromRequest.dl_primitive 		= DL_PROMISCON_REQ;
  	PromRequest.dl_level 			= level;

        ctlbuf.len = sizeof(PromRequest);
        ctlbuf.buf = (caddr_t)&PromRequest;

        if (putmsg(fd, &ctlbuf ,NULL,0)  < 0) {
                perror("Prom Mode:");
                return(0);
        }

        ctlbuf.maxlen = sizeof(union DL_primitives);
        ctlbuf.len = 0;
        ctlbuf.buf = (char *)&rcvbuf;
        if ((retval = getmsg(fd, &ctlbuf ,NULL, &flags))  < 0) {
                perror("Prom Mode ack!");
                return(0);
        }

	return (Acknoledge(&rcvbuf,DL_OK_ACK,"DL_PROMISCON_REQ"));
}
#endif /* SOLARIS2 */
